program ecology;
{     This program simulates the ecology in a strange universe.  The universe
  contains white, black, creatures that eat white, that eat black, that live
  on white and eat other creatures, and creatures that live on black and eat
  other creatures.  These beasts are called whiteherb, blackherb, whitecarn,
  blackcarn respectively.  Carn stands for carnivore and herb stands for
  herbivore.  The color represents the color of the terrain the creature lives
  on A CREATURE_REC node is created for each creature stored in a circular
  list.
      The environment is the screen.  The terrain, white and black, represent
  two types of sustinance for herbivores, almost but not quite like vegitation.
  Black and white do not grow but are created by the herbivores.  When a
  WHITEHERB eats a piece of black it leaves a piece of white behind.  The
  byte for each cell on the screen stores the terrain type.
      The program simulates the lives of these creatures and there
  environment.  It consists of a loop over the creature list.  At each
  iteration the CURRENT creature gets to move, eat, and reproduce.  Since the
  list is circular the loop never ends and the beasts get to move in turn. }



const
  xlimit = 80;
  ylimit = 48;
  xsize = 40;
  ysize = 24;
  blank = 32;
  screen = $b800;

{$U-}
{$C-}


type
  creature_ty = (whiteherb, blackherb, whitecarn, blackcarn, whitesp, blacksp);
  dir_ty = (north, east, south, west);

  creature_ptr = ^creature_rec;

  creature_rec = record
    creature: creature_ty;
    xpos, ypos, strength: integer;
    dir: dir_ty;
    next, last: creature_ptr;
  end;


var
  gen,                       { Generation number. }
  i: integer;                { Loop variable. }
  top,                       { Points to first creature in loop. }
  current: creature_ptr;     { Points to current creature. }
  xorig, yorig: integer;     { Location of screen origin in CTRLINK. }

  { Environment description variables. }
  blackcell, whitecell: creature_ptr;
  ctrcolor: array[whiteherb..blacksp] of integer;
  ctrback:  array[whiteherb..blacksp] of integer;
  ctrshape: array[whiteherb..blacksp, north..west] of char;
  ctrlink: array[1..xlimit, 1..ylimit] of creature_ptr;

  { Miscellanious variables. }
  graph: text;
  line: string[80];


procedure drawsc;

  var
    x, y: integer;

  begin
    for y:=1 to ysize do
    begin
      gotoxy(1, y);

      for x:=1 to xsize do
        with ctrlink[x+xorig-1, y+yorig-1]^ do
        begin
          textcolor(ctrcolor[creature]);
          textbackground(ctrback[creature]);
          write(ctrshape[creature, dir]);
        end;
    end;
  end;   { of the procedure drawsc. }


procedure draw(current: creature_ptr; xp, yp: integer; moved: boolean);

  begin
    with current^ do
    begin
      if (xp>=xorig) and (xp<xorig+xsize) and
         (yp>=yorig) and (yp<yorig+ysize) then
      begin
        textcolor(ctrcolor[creature]);
        textbackground(ctrback[creature]);
        gotoxy(xp-xorig+1, yp-yorig+1);
        write(ctrshape[creature, dir]);
      end;

      if moved then
      begin
        if (xpos>=xorig) and (xpos<xorig+xsize) and
           (ypos>=yorig) and (ypos<yorig+ysize) then
        begin
          textbackground(ctrback[creature]);
          gotoxy(xpos-xorig+1, ypos-yorig+1);
          write(' ');
        end;
        if odd(integer(creature)) then
          ctrlink[xpos, ypos]:=blackcell
        else
          ctrlink[xpos, ypos]:=whitecell;
        xpos:=xp;
        ypos:=yp;
        ctrlink[xp, yp]:=current;
      end;
    end;
  end;   { of the procedure draw. }


function emptysp(x, y: integer): boolean;

  begin
    emptysp := (ctrlink[x, y]=whitecell) or (ctrlink[x, y]=blackcell);
  end;   { of the function emptysp. }


function looksp(x, y: integer): creature_ty;

  begin
    looksp:=ctrlink[x, y]^.creature;
  end;   { of the function looksp. }


procedure clearsp(x, y: integer);

  begin
    with ctrlink[x, y]^ do
      if (xpos>=xorig) and (xpos<xorig+xsize) and
         (ypos>=yorig) and (ypos<yorig+ysize) then
      begin
        textbackground(ctrback[creature]);
        gotoxy(xpos-xorig+1, ypos-yorig+1);
        write(' ');
      end;

    if odd(integer(ctrlink[x, y]^.creature)) then
      ctrlink[x, y] := blackcell
    else
      ctrlink[x, y] := whitecell;
  end;   { of the procedure clearsp. }


procedure initialize(var current: creature_ptr);
{ This procedure initializes the creature list and the environment. }

  var
    x, y: integer;

  begin
    textmode(c40);
    xorig := 1;
    yorig := 1;

    new(blackcell);
    with blackcell^ do
    begin
      creature := blacksp;
      dir := north;
      last:=nil;
      next:=nil;
      xpos:=0;
      ypos:=0;
      strength:=0;
    end;

    new(whitecell);
    with whitecell^ do
    begin
      creature := whitesp;
      dir := north;
      last:=nil;
      next:=nil;
      xpos:=0;
      ypos:=0;
      strength:=0;
    end;

    { Initialize the creature shape and color arrays. }
    ctrcolor[whiteherb]:=lightblue;
    ctrback[whiteherb]:=magenta;
    ctrshape[whiteherb, north]:=chr(202);
    ctrshape[whiteherb, south]:=chr(203);
    ctrshape[whiteherb, east]:=chr(204);
    ctrshape[whiteherb, west]:=chr(185);

    ctrcolor[blackherb]:=green;
    ctrback[blackherb]:=black;
    ctrshape[blackherb, north]:=chr(202);
    ctrshape[blackherb, south]:=chr(203);
    ctrshape[blackherb, east]:=chr(204);
    ctrshape[blackherb, west]:=chr(185);

    ctrcolor[whitecarn]:=lightred;
    ctrback[whitecarn]:=magenta;
    ctrshape[whitecarn, north]:=chr(24);
    ctrshape[whitecarn, south]:=chr(25);
    ctrshape[whitecarn, east]:=chr(26);
    ctrshape[whitecarn, west]:=chr(27);

    ctrcolor[blackcarn]:=red;
    ctrback[blackcarn]:=black;
    ctrshape[blackcarn, north]:=chr(24);
    ctrshape[blackcarn, south]:=chr(25);
    ctrshape[blackcarn, east]:=chr(26);
    ctrshape[blackcarn, west]:=chr(27);

    ctrcolor[whitesp] := white;
    ctrback[whitesp] := magenta;
    ctrshape[whitesp, north] := ' ';
    ctrshape[whitesp, south] := ' ';
    ctrshape[whitesp, east] := ' ';
    ctrshape[whitesp, west] := ' ';

    ctrcolor[blacksp] := white;
    ctrback[blacksp] := black;
    ctrshape[blacksp, north] := ' ';
    ctrshape[blacksp, south] := ' ';
    ctrshape[blacksp, east] := ' ';
    ctrshape[blacksp, west] := ' ';

    { Initialize the terrain. }
    for x:=1 to xlimit do
      for y:=1 to ylimit do
      begin

        if random(2)=0 then
        begin
          ctrlink[x, y] := whitecell;
          textbackground(magenta);
        end
        else
        begin
          ctrlink[x, y] := blackcell;
          textbackground(black);
        end;

      end;   { of the loop to set the terrain. }

    drawsc;

    { Create the creature list and draw the creatures on the screen. }
    top:=nil;
    for x:=1 to 240 do
    begin

      if top=nil then
      begin
        new(top);
        current:=top;
      end
      else
      begin
        new(current^.next);
        current^.next^.last:=current;
        current:=current^.next;
      end;

      with current^ do
      begin
        repeat
          xpos:=random(xlimit)+1;
          ypos:=random(ylimit)+1;
        until emptysp(xpos, ypos);
        y:=random(40);
        if y>37 then
        begin
          creature:=creature_ty(y-36);
          strength:=150;
        end
        else
        begin
          creature:=creature_ty(y div 19);
          strength:=40;
        end;
        dir:=dir_ty(random(4));
        draw(current, xpos, ypos, false);
      end;

    end;   { of the loop to create creatures. }

    { Close the ring. }
    current^.next:=top;
    top^.last:=current;

  end;   { of the procedure initialize. }


function turn(dir: dir_ty; whichway: integer): dir_ty;

  begin
    if whichway=0 then
      if dir=north then
        turn:=west
      else
        turn:=pred(dir)
    else
      if dir=west then
        turn:=north
      else
        turn:=succ(dir);
  end;   { of the function turn. }


procedure newpos(var xpos, ypos, xp, yp: integer; dir: dir_ty);

  begin
    xp:=xpos;
    yp:=ypos;
    case dir of
      north: if yp=1 then yp:=ylimit else yp:=yp-1;
      south: if yp=ylimit then yp:=1 else yp:=yp+1;
      west: if xp=1 then xp:=xlimit else xp:=xp-1;
      east: if xp=xlimit then xp:=1 else xp:=xp+1;
    end;
  end;   { of the procedure newpos. }


function look(var xpos, ypos, xp, yp: integer; dir: dir_ty;
              creature: creature_ty): boolean;

  begin
    newpos(xpos, ypos, xp, yp, dir);
    if emptysp(xp, yp) then
      if creature>blackherb then
        if (odd(integer(creature)) =
           odd(integer(looksp(xp, yp)))) then
          look:=true
        else
          look:=false
      else
        if (odd(integer(creature)) <>
           odd(integer(looksp(xp, yp)))) then
          look:=true
        else
          look:=false
    else
      look:=false;
  end;   { of the function look. }


function prey(var xpos, ypos, xp, yp: integer; dir: dir_ty;
              creature: creature_ty): boolean;

  var
    i: dir_ty;

  begin
    newpos(xpos, ypos, xp, yp, dir);
    if emptysp(xp, yp) then
      prey:=false
    else if looksp(xp, yp)<>creature then
      prey:=true
    else
      prey:=false;
  end;   { of the function prey. }


procedure starve(var current: creature_ptr);

  var
    temp: creature_ptr;

  begin
    with current^ do
    begin

      { Has the creature starved. }
      strength:=strength-1;
      if strength<0 then

        { If this is the last creature in the loop then the loop disappears. }
        if current^.next=current then
        begin
          dispose(current);
          current:=nil;
        end

        else
        begin
          clearsp(xpos, ypos);
          temp:=current;
          current:=current^.last;
          current^.next:=temp^.next;
          temp^.next^.last:=current;
          if top=temp then top:=temp^.next;
          dispose(temp);
        end;

    end;
  end;   { of the procedure starve. }


procedure reproduce(var current: creature_ptr);

  var
    startdir, newdir: dir_ty;
    xp, yp, i: integer;
    done: boolean;
    temp: creature_ptr;

  begin
    with current^ do
    begin

      { Should the creature reproduce. }
      if ((creature<whitecarn) and (strength>80)) or
         ((creature>blackherb) and (strength>500)) then
      begin
        startdir:=turn(turn(dir, 0), 0);
        newdir:=startdir;
        i:=random(2);
        done:=false;

        repeat
          newpos(xpos, ypos, xp, yp, newdir);

          if emptysp(xp, yp) then
          begin
            new(temp);
            current^.last^.next:=temp;
            temp^.last:=current^.last;
            temp^.next:=current;
            current^.last:=temp;
            temp^.creature:=creature;
            temp^.xpos:=xp;
            temp^.ypos:=yp;
            temp^.dir:=newdir;
            draw(temp, xp, yp, false);

            if creature<whitecarn then
            begin
              current^.strength:=current^.strength-40;
              temp^.strength:=35;
            end
            else
            begin
              current^.strength:=current^.strength-350;
              temp^.strength:=125;
            end;
            done:=true;
          end
          else
            newdir:=turn(newdir, i);
        until done or (newdir=startdir);
      end;

    end;
  end;   { of the procedure reproduce. }


procedure doherb(var current: creature_ptr);

  label
    pickturn;

  var
    holdx, holdy, i, xp, yp: integer;

  begin
    with current^ do
    begin

      { Should the creature go forward. }
      if random(16)<>0 then
        if look(xpos, ypos, xp, yp, dir, creature) then
        begin
          draw(current, xp, yp, true);
          strength:=strength+2;
        end
        else
          goto pickturn
      else

      { Should the creature turn. }
      begin
pickturn:
        i:=random(2);
        newpos(xpos, ypos, holdx, holdy, dir);
        if look(xpos, ypos, xp, yp, turn(dir, i), creature) then
        begin
          dir:=turn(dir, i);
          draw(current, xpos, ypos, false);
        end
        else if look(xpos, ypos, xp, yp, turn(dir, (i+1) mod 2), creature) then
        begin
          dir:=turn(dir, (i+1) mod 2);
          draw(current, xpos, ypos, false);
        end
        else if not emptysp(holdx, holdy) or (random(8)=0) then
        begin
          dir:=turn(dir, i);
          draw(current, xpos, ypos, false);
        end

        { It is unlikely that there is anything to eat nearby so lets go
          forward. }
        else
        begin
          if odd(integer(creature)) <> odd(integer(looksp(holdx, holdy))) then
            strength:=strength+2;
          draw(current, holdx, holdy, true);
        end;
      end;

      { Should the creature reproduce. }
      reproduce(current);

      { Decrease strength. }
      starve(current);

    end;
  end;   { of the procedure doherb. }


procedure eat(var carnstr, xp, yp: integer);

  var
    temp: creature_ptr;

  begin

    if emptysp(xp, yp) then
    begin
      gotoxy(1, 25);
      write('Eat error: ', xp, ', ', yp);
      read(line);
    end

    else
      with ctrlink[xp, yp]^ do
      begin
        temp:=ctrlink[xp, yp];
        last^.next:=next;
        next^.last:=last;
        if top=ctrlink[xp, yp] then top:=next;
        carnstr:=carnstr+10+strength div 4;
        clearsp(xp, yp);
        dispose(temp);
      end;

  end;   { of the procedure eat. }


procedure docarn(var current: creature_ptr);

  label
    pickturn;

  var
    xp, yp, i: integer;

  begin
    with current^ do
    begin

      { See if there is any prey nearby. }
      i:=random(2);
      if prey(xpos, ypos, xp, yp, dir, creature) then
      begin
        eat(strength, xp, yp);
        if odd(integer(creature)) = odd(integer(looksp(xp, yp))) then
          draw(current, xp, yp, true)
        else
          goto pickturn;
      end

      else if prey(xpos, ypos, xp, yp, turn(dir, i), creature) then
      begin
        eat(strength, xp, yp);
        dir:=turn(dir, i);
        draw(current, xpos, ypos, false);
      end

      else if prey(xpos, ypos, xp, yp, turn(dir, (i+1) mod 2), creature) then
      begin
        eat(strength, xp, yp);
        dir:=turn(dir, (i+1) mod 2);
        draw(current, xpos, ypos, false);
      end

      { Try to go forward. }
      else if random(8)<>0 then
        if look(xpos, ypos, xp, yp, dir, creature) then
          draw(current, xp, yp, true)
        else
          goto pickturn
      else

      { Should the creature turn. }
      begin
pickturn:
        i:=random(2);
        if look(xpos, ypos, xp, yp, turn(dir, i), creature) then
        begin
          dir:=turn(dir, i);
          draw(current, xpos, ypos, false);
        end
        else
        begin
          dir:=turn(dir, (i+1) mod 2);
          draw(current, xpos, ypos, false);
        end;
      end;

      reproduce(current);

      { Decrease strength. }
      starve(current);

    end;
  end;   { of the procedure docarn. }


procedure countring;

  var
    countwh, countbh, countwc, countbc: integer;
    temp: creature_ptr;

  begin
    countwh:=0;
    countbh:=0;
    countwc:=0;
    countbc:=0;
    temp:=top;
    gen:=gen+1;

    repeat
      case temp^.creature of
        whiteherb: countwh:=countwh+1;
        blackherb: countbh:=countbh+1;
        whitecarn: countwc:=countwc+1;
        blackcarn: countbc:=countbc+1;
      end;
      temp:=temp^.next;
    until temp=top;

    gotoxy(1, 25);
    write('wh: ', countwh, ' bh: ', countbh,
          ' wc: ', countwc, ' bc: ', countbc, '     ');
    writeln(graph, gen, ', ', countwh, ', ', countbh, ', ',
            countwc, ', ', countbc);
  end;   { of the procedure countring. }



begin
  writeln('Enter the file name.');
  readln(line);
  assign(graph, line);
  rewrite(graph);
  gen:=0;
  initialize(current);

  while (current<>nil) and not keypressed do
    with current^ do
    begin

      case creature of
        whiteherb, blackherb: doherb(current);
        whitecarn, blackcarn: docarn(current);
      end;

      if current<>nil then
        current:=current^.next;
      if current=top then countring;
    end;   { of the loop to move a creature. }

  readln(line);
  textmode(bw80);
  close(graph);
end.